import inspect


class State(object):
    EVENT_ENTER = 'enter'
    EVENT_UPDATE = 'update'
    EVENT_EXIT = 'exit'

    def __init__(self, default=None):
        self.default = default if default is not None else False
        self.__event_handlers = {}
        self.__message_handlers = {}

    def _on_event(self, event):
        def decorator(f):
            self.__event_handlers[event] = f.__name__
            return f
        return decorator

    @property
    def on_enter(self):
        return self._on_event(self.EVENT_ENTER)

    @property
    def on_update(self):
        return self._on_event(self.EVENT_UPDATE)

    @property
    def on_exit(self):
        return self._on_event(self.EVENT_EXIT)

    def on_message(self, message_name):
        def decorator(f):
            self.__message_handlers[message_name] = f.__name__
            return f
        return decorator

    def _handle_event(self, event, state_machine):
        if event in self.__event_handlers:
            function_name = self.__event_handlers[event]
            getattr(state_machine, function_name)()

    def handle_enter(self, state_machine):
        self._handle_event(self.EVENT_ENTER, state_machine)

    def handle_update(self, state_machine):
        self._handle_event(self.EVENT_UPDATE, state_machine)

    def handle_exit(self, state_machine):
        self._handle_event(self.EVENT_EXIT, state_machine)

    def handle_message(self, state_machine, message_name, *args, **kwargs):
        if message_name in self.__message_handlers:
            function_name = self.__message_handlers[message_name]
            getattr(state_machine, function_name)(*args, **kwargs)


class StateMachine(object):
    def __init__(self):
        states = inspect.getmembers(self, lambda m: isinstance(m, State))
        self.__states = dict(states)
        default_states = list(filter(lambda s: s[1].default, states))
        if len(default_states) != 1:
            raise Exception('Must have exactly 1 default state.')
        self.state = default_states[0][1]

    def send_message(self, message_name, *args, **kwargs):
        self.state.handle_message(self, message_name, *args, **kwargs)

    def update(self):
        self.state.handle_update(self)

    def set_state(self, state):
        changed = self.state != state
        if changed:
            self.state.handle_exit(self)
        self.state = state
        if changed:
            self.state.handle_enter(self)



from setuptools import setup
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
        long_description = f.read()

setup(
    name='finesm',

    version='0.0.0',

    description='A fine state machine',
    long_description=long_description,

    url='https://github.com/wesleyks/finesm',

    license='MIT',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
    ],

    keywords='finite-state machine state',

    packages=['finesm'],
)
